package util;

// Java imports
import java.io.*;

// Little-endian implementation of DataInputStream
public class LEDataInputStream extends FilterInputStream {

    // Instance fields
    private DataInputStream data; // Data-decoding stream



    ///////////////////////////////////////////////////////////////////////////
    //                             Constructors                              //
    ///////////////////////////////////////////////////////////////////////////

    // Default constructor
    public LEDataInputStream(InputStream stream) {
        super(stream = stream instanceof LEDataInputStream ?
            ((LEDataInputStream) stream).in : stream);
        data = stream instanceof DataInputStream ?
             (DataInputStream) stream : new DataInputStream(stream);
    }

    // Byte array constructor
    public LEDataInputStream(byte[] data) {
        this(new ByteArrayInputStream(data));
    }



    ///////////////////////////////////////////////////////////////////////////
    //                            Public Methods                             //
    ///////////////////////////////////////////////////////////////////////////

    // Reads a boolean from the underlying input stream
    public boolean readBoolean() throws IOException {
        return data.readByte() != 0;
    }

    // Reads an array of booleans from the underlying input stream
    public boolean[] readBooleans(int count) throws IOException {
        return readBooleans(new boolean[count], 0, count);
    }

    // Reads an array of booleans from the underlying input stream
    public boolean[] readBooleans(boolean[] v, int offset, int length)
        throws IOException {
        for (int x = 0; x < length; x++)
            v[offset + x] = readBoolean();
        return v;
    }

    // Reads a byte from the underlying input stream
    public byte readByte() throws IOException {
        return data.readByte();
    }

    // Reads an array of bytes from the underlying input stream
    public byte[] readBytes(int count) throws IOException {
        return readBytes(new byte[count], 0, count);
    }

    // Reads an array of bytes from the underlying input stream
    public byte[] readBytes(byte[] v, int offset, int length)
        throws IOException {
        data.read(v, offset, length);
        return v;
    }

    // Reads a character from the underlying input stream
    public char readChar() throws IOException {
        return (char) Short.reverseBytes(data.readShort());
    }

    // Reads an array of characters from the underlying input stream
    public char[] readChars(int count) throws IOException {
        return readChars(new char[count], 0, count);
    }

    // Reads an array of characters from the underlying input stream
    public char[] readChars(char[] v, int offset, int length)
        throws IOException {
        for (int x = 0; x < length; x++)
            v[offset + x] = readChar();
        return v;
    }

    // Reads a double from the underlying input stream
    public double readDouble() throws IOException {
        return readDouble(false);
    }

    // Reads a double from the underlying input stream
    public double readDouble(boolean finite) throws IOException {
        long   bits = readLong();
        double ret  = Double.longBitsToDouble(bits);
        if (finite && !Double.isFinite(ret)) throw new RuntimeException(
            String.format("Non-finite double value 0x%016X", bits));
        return ret;
    }

    // Reads an array of doubles from the underlying input stream
    public double[] readDoubles(int count) throws IOException {
        return readDoubles(new double[count], 0, count, false);
    }

    // Reads an array of doubles from the underlying input stream
    public double[] readDoubles(int count, boolean finite) throws IOException {
        return readDoubles(new double[count], 0, count, finite);
    }

    // Reads an array of doubles from the underlying input stream
    public double[] readDoubles(double[] v, int offset, int length)
        throws IOException {
        return readDoubles(v, offset, length, false);
    }

    // Reads an array of doubles from the underlying input stream
    public double[] readDoubles(double[] v, int offset, int length,
        boolean finite) throws IOException {
        for (int x = 0; x < length; x++)
            v[offset + x] = readDouble(finite);
        return v;
    }

    // Reads a float from the underlying input stream
    public float readFloat() throws IOException {
        return readFloat(false);
    }

    // Reads a float from the underlying input stream
    public float readFloat(boolean finite) throws IOException {
        int   bits = readInt();
        float ret  = Float.intBitsToFloat(bits);
        if (finite && !Float.isFinite(ret)) throw new RuntimeException(
            String.format("Non-finite float value 0x%08X", bits));
        return ret;
    }

    // Reads an array of floats from the underlying input stream
    public float[] readFloats(int count) throws IOException {
        return readFloats(new float[count], 0, count, false);
    }

    // Reads an array of floats from the underlying input stream
    public float[] readFloats(int count, boolean finite) throws IOException {
        return readFloats(new float[count], 0, count, finite);
    }

    // Reads an array of floats from the underlying input stream
    public float[] readFloats(float[] v, int offset, int length)
        throws IOException {
        return readFloats(v, offset, length, false);
    }

    // Reads an array of floats from the underlying input stream
    public float[] readFloats(float[] v, int offset, int length,
        boolean finite) throws IOException {
        for (int x = 0; x < length; x++)
            v[offset + x] = readFloat(finite);
        return v;
    }

    // Reads an int from the underlying input stream
    public int readInt() throws IOException {
        return Integer.reverseBytes(data.readInt());
    }

    // Reads an array of ints from the underlying input stream
    public int[] readInts(int count) throws IOException {
        return readInts(new int[count], 0, count);
    }

    // Reads an array of ints from the underlying input stream
    public int[] readInts(int[] v, int offset, int length)
        throws IOException {
        for (int x = 0; x < length; x++)
            v[offset + x] = readInt();
        return v;
    }

    // Reads a long from the underlying input stream
    public long readLong() throws IOException {
        return Long.reverseBytes(data.readLong());
    }

    // Reads an array of longs from the underlying input stream
    public long[] readLongs(int count) throws IOException {
        return readLongs(new long[count], 0, count);
    }

    // Reads an array of longs from the underlying input stream
    public long[] readLongs(long[] v, int offset, int length)
        throws IOException {
        for (int x = 0; x < length; x++)
            v[offset + x] = readLong();
        return v;
    }

    // Reads a 24-bit integer from the underlying input stream
    public int readMiddle() throws IOException {
        int v = readShort() & 0xFFFF;
        return readByte() << 16 | v;
    }

    // Reads an array of 24-bit integers from the underlying input stream
    public int[] readMiddles(int count) throws IOException {
        return readMiddles(new int[count], 0, count);
    }

    // Reads an array of 24-bit integers from the underlying input stream
    public int[] readMiddles(int[] v, int offset, int length)
        throws IOException {
        for (int x = 0; x < length; x++)
            v[offset + x] = readMiddle();
        return v;
    }

    // Reads a short from the underlying input stream
    public short readShort() throws IOException {
        return Short.reverseBytes(data.readShort());
    }

    // Reads an array of shorts from the underlying input stream
    public short[] readShorts(int count) throws IOException {
        return readShorts(new short[count], 0, count);
    }

    // Reads an array of shorts from the underlying input stream
    public short[] readShorts(short[] v, int offset, int length)
        throws IOException {
        for (int x = 0; x < length; x++)
            v[offset + x] = readShort();
        return v;
    }

    // Reads from the stream in a modified UTF-8 string
    public String readUTF() throws IOException {
        return data.readUTF();
    }

    // Advances forward in the stream some number of bytes
    public int skipBytes(int n) throws IOException {
        return data.skipBytes(n);
    }

}
