package util;

// Java imports
import java.io.*;
import java.nio.charset.*;

// Little-endian implementation of DataOutputStream
public class LEDataOutputStream extends FilterOutputStream {

    // Instance fields
    private DataOutputStream data; // Data-encoding stream



    ///////////////////////////////////////////////////////////////////////////
    //                             Constructors                              //
    ///////////////////////////////////////////////////////////////////////////

    // Default constructor
    public LEDataOutputStream() {
        this(new ByteArrayOutputStream());
    }

    // Stream constructor
    public LEDataOutputStream(OutputStream stream) {
        super(stream = stream instanceof LEDataOutputStream ?
            ((LEDataOutputStream) stream).out : stream);
        data = stream instanceof DataOutputStream ?
             (DataOutputStream) stream : new DataOutputStream(stream);
    }



    ///////////////////////////////////////////////////////////////////////////
    //                            Public Methods                             //
    ///////////////////////////////////////////////////////////////////////////

    // Returns the current value of the underlying buffer
    public int size() {
        return out instanceof ByteArrayOutputStream ?
            ((ByteArrayOutputStream) out).size() : -1;
    }

    // Produce a byte array containing this stream's data
    public byte[] toByteArray() throws IOException {
        return out instanceof ByteArrayOutputStream ?
            ((ByteArrayOutputStream) out).toByteArray() : null;
    }

    // Writes a boolean to the underlying output stream as a 1-byte value
    public void writeBoolean(boolean v) throws IOException {
        data.writeBoolean(v);
    }

    // Writes an array of booleans to the underlying output stream
    public void writeBooleans(boolean[] v) throws IOException {
        writeBooleans(v, 0, v.length);
    }

    // Writes an array of booleans to the underlying output stream
    public void writeBooleans(boolean[] v, int offset, int length)
    throws IOException {
        for (int x = 0; x < length; x++)
            writeBoolean(v[offset + x]);
    }

    // Writes out a byte to the underlying output stream as a 1-byte value
    public void writeByte(int v) throws IOException {
        data.writeByte(v);
    }

    // Writes out the string to the underlying output stream
    public void writeBytes(String s) throws IOException {
        data.writeBytes(s);
    }

    // Writes an array of bytes to the underlying output stream
    public void writeBytes(byte[] v) throws IOException {
        writeBytes(v, 0, v.length);
    }

    // Writes an array of bytes to the underlying output stream
    public void writeBytes(byte[] v, int offset, int length)
    throws IOException {
        data.write(v, offset, length);
    }

    // Writes a character to the underlying output stream as a 2-byte value
    public void writeChar(int v) throws IOException {
        data.writeChar(Short.reverseBytes((short) v));
    }

    // Writes a string to the underlying output stream
    public void writeChars(String s) throws IOException {
        writeChars(s.toCharArray());
    }

    // Writes an array of characters to the underlying output stream
    public void writeChars(char[] v) throws IOException {
        writeChars(v, 0, v.length);
    }

    // Writes an array of characters to the underlying output stream
    public void writeChars(char[] v, int offset, int length)
    throws IOException {
        for (int x = 0; x < length; x++)
            writeChar(v[offset + x]);
    }

    // Converts the double argument to a long using doubleToLongBits
    public void writeDouble(double v) throws IOException {
        data.writeLong(Long.reverseBytes(Double.doubleToLongBits(v)));
    }

    // Writes an array of doubles to the underlying output stream
    public void writeDoubles(double[] v) throws IOException {
        writeDoubles(v, 0, v.length);
    }

    // Writes an array of doubles to the underlying output stream
    public void writeDoubles(double[] v, int offset, int length)
    throws IOException {
        for (int x = 0; x < length; x++)
            writeDouble(v[offset + x]);
    }

    // Converts the float argument to an int using floatToIntBits
    public void writeFloat(float v) throws IOException {
        data.writeInt(Integer.reverseBytes(Float.floatToIntBits(v)));
    }

    // Writes an array of floats to the underlying output stream
    public void writeFloats(float[] v) throws IOException {
        writeFloats(v, 0, v.length);
    }

    // Writes an array of floats to the underlying output stream
    public void writeFloats(float[] v, int offset, int length)
    throws IOException {
        for (int x = 0; x < length; x++)
            writeFloat(v[offset + x]);
    }

    // Writes an int to the underlying output stream as four bytes
    public void writeInt(int v) throws IOException {
        data.writeInt(Integer.reverseBytes(v));
    }

    // Writes an array of ints to the underlying output stream
    public void writeInts(int[] v) throws IOException {
        writeInts(v, 0, v.length);
    }

    // Writes an array of ints to the underlying output stream
    public void writeInts(int[] v, int offset, int length)
    throws IOException {
        for (int x = 0; x < length; x++)
            writeInt(v[offset + x]);
    }

    // Writes an int to the underlying output stream as three bytes
    public void writeMiddle(int v) throws IOException {
        writeShort((short)  v       );
        writeByte ((byte ) (v >> 16));
    }

    // Writes an array of 24-bit integers to the underlying output stream
    public void writeMiddles(int[] v) throws IOException {
        writeMiddles(v, 0, v.length);
    }

    // Writes an array of 24-bit integers to the underlying output stream
    public void writeMiddles(int[] v, int offset, int length)
    throws IOException {
        for (int x = 0; x < length; x++)
            writeMiddle(v[offset + x]);
    }

    // Writes a long to the underlying output stream as eight bytes
    public void writeLong(long v) throws IOException {
        data.writeLong(Long.reverseBytes(v));
    }

    // Writes an array of longs to the underlying output stream
    public void writeLong(long[] v) throws IOException {
        writeLongs(v, 0, v.length);
    }

    // Writes an array of longs to the underlying output stream
    public void writeLongs(long[] v, int offset, int length)
    throws IOException {
        for (int x = 0; x < length; x++)
            writeLong(v[offset + x]);
    }

    // Writes a short to the underlying output stream as two bytes
    public void writeShort(int v) throws IOException {
        data.writeShort(Short.reverseBytes((short) v));
    }

    // Writes an array of shorts to the underlying output stream
    public void writeShorts(short[] v) throws IOException {
        writeShorts(v, 0, v.length);
    }

    // Writes an array of shorts to the underlying output stream
    public void writeShorts(short[] v, int offset, int length)
    throws IOException {
        for (int x = 0; x < length; x++)
            writeShort(v[offset + x]);
    }

    // Writes a string to the underlying output stream using modified UTF-8
    public void writeUTF(String str) throws IOException {
        data.writeUTF(str);
    }

}
