// This file is included through NativeVUE.c and cannot be built directly. */
#ifdef NATIVEVUE



///////////////////////////////////////////////////////////////////////////////
//                                 Constants                                 //
///////////////////////////////////////////////////////////////////////////////

// Token types
#define BINARY 4
#define FLOAT  1
#define UNARY  3
#define SYMBOL 2
#define WORD   0

// CPU stages
#define EXCEPTION 2
#define EXECUTE   1

// Operator IDs
#define ADD                     2
#define BITWISE_AND             3
#define BITWISE_NOT             4
#define BITWISE_OR              5
#define BITWISE_XOR             6
#define CEIL                    7
#define DIVIDE                  8
#define EQUAL                   9
#define FLOOR                  10
//#define FLOAT                 1 // Same as type
#define GREATER_EQUAL_SIGNED   11
#define GREATER_EQUAL_UNSIGNED 12
#define GREATER_SIGNED         13
#define GREATER_UNSIGNED       14
#define GROUP                  15
#define LESS_EQUAL_SIGNED      16
#define LESS_EQUAL_UNSIGNED    17
#define LESS_SIGNED            18
#define LESS_UNSIGNED          19
#define LOGICAL_AND            20
#define LOGICAL_NOT            21
#define LOGICAL_OR             22
#define LOGICAL_XOR            23
#define MULTIPLY               24
#define NEGATE                 25
#define NOT_EQUAL              26
#define READ_WORD              27
#define REMAINDER              28
#define ROUND                  29
#define SHIFT_LEFT             30
#define SHIFT_RIGHT            31
#define SHIFT_RIGHT_ARITHMETIC 32
#define SUBTRACT               33
#define TRUNC                  34
//#define WORD                  0 // Same as type
#define XFLOAT                 35
#define XWORD                  36

// Functional symbol IDs
#define ADDRESS    0
#define CODE       1
#define COND       2
#define DISP       3
#define FORMAT     4
#define ID         5
#define IMM        6
#define OPCODE     7
#define REG1       8
#define REG2       9
#define REGID     10
#define SIZE      11
#define SUBOPCODE 12
#define VALUE     13
#define VECTOR    14



///////////////////////////////////////////////////////////////////////////////
//                            Evaluation Helpers                             //
///////////////////////////////////////////////////////////////////////////////

// Resolve a float from word bits
static float asFloat(int32_t value) {
    return *(float *)&value;
}

// Resolve the word bits of a float
static int32_t asWord(float value) {
    int32_t bits   = *(int32_t *)&value;
    int32_t exp    = bits & 0x7F800000;
    int32_t digits = bits & 0x007FFFFF;
    return
        !(bits & 0x7FFFFFFF) || // Zero
        exp   == 0x7F800000  || // Indefinite
        !exp  && digits != 0    // Denormal
    ? 0 : bits;
}

// Convert a float to a word
static int32_t toWord(float value) {
    value = roundf(value);
    return value>=0x7FFFFFFF || value<(int32_t)0x80000000 ? 0 : (int) value;
}



///////////////////////////////////////////////////////////////////////////////
//                             Binary Functions                              //
///////////////////////////////////////////////////////////////////////////////

// Evaluate +
static int32_t evalAdd(
    int32_t leftWord,int32_t leftValue,int32_t rightWord,int32_t rightValue) {
    return leftWord && rightWord ? leftValue + rightValue : asWord(
        (leftWord  ? (float) leftValue  : asFloat(leftValue )) +
        (rightWord ? (float) rightValue : asFloat(rightValue))
    );
}

// Evaluate &
static int32_t evalBitwiseAnd(
    int32_t leftWord,int32_t leftValue,int32_t rightWord,int32_t rightValue) {
    return (leftWord  ? leftValue  : toWord(asFloat(leftValue ))) &
           (rightWord ? rightValue : toWord(asFloat(rightValue)));
}

// Evaluate |
static int32_t evalBitwiseOr(
    int32_t leftWord,int32_t leftValue,int32_t rightWord,int32_t rightValue) {
    return (leftWord  ? leftValue  : toWord(asFloat(leftValue ))) |
           (rightWord ? rightValue : toWord(asFloat(rightValue)));
}

// Evaluate ^
static int32_t evalBitwiseXOr(
    int32_t leftWord,int32_t leftValue,int32_t rightWord,int32_t rightValue) {
    return (leftWord  ? leftValue  : toWord(asFloat(leftValue ))) ^
           (rightWord ? rightValue : toWord(asFloat(rightValue)));
}

// Evaluate /
static int32_t evalDivide(
    int32_t leftWord,int32_t leftValue,int32_t rightWord,int32_t rightValue) {
    return !(rightWord ? rightValue : rightValue & 0x7FFFFFFF) ? 0 :
    leftWord && rightWord ? leftValue / rightValue : asWord(
        (leftWord  ? (float) leftValue  : asFloat(leftValue )) /
        (rightWord ? (float) rightValue : asFloat(rightValue))
    );
}

// Evaluate ==
static int32_t evalEqual(
    int32_t leftWord,int32_t leftValue,int32_t rightWord,int32_t rightValue) {
    return leftWord && rightWord ? leftValue == rightValue : (
        (leftWord  ? (float) leftValue  : asFloat(leftValue )) ==
        (rightWord ? (float) rightValue : asFloat(rightValue))
    );
}

// Evaluate >=
static int32_t evalGreaterEqualSigned(
    int32_t leftWord,int32_t leftValue,int32_t rightWord,int32_t rightValue) {
    return leftWord && rightWord ? leftValue >= rightValue : (
        (leftWord  ? (float) leftValue  : asFloat(leftValue )) >=
        (rightWord ? (float) rightValue : asFloat(rightValue))
    );
}

// Evaluate >=_
static int32_t evalGreaterEqualUnsigned(
    int32_t leftWord,uint32_t leftValue,int32_t rightWord,uint32_t rightValue){
    return leftWord && rightWord ? leftValue >= rightValue : (
        (leftWord  ? (float) leftValue  : asFloat(leftValue )) >=
        (rightWord ? (float) rightValue : asFloat(rightValue))
    );
}

// Evaluate >
static int32_t evalGreaterSigned(
    int32_t leftWord,int32_t leftValue,int32_t rightWord,int32_t rightValue) {
    return leftWord && rightWord ? leftValue > rightValue : (
        (leftWord  ? (float) leftValue  : asFloat(leftValue )) >
        (rightWord ? (float) rightValue : asFloat(rightValue))
    );
}

// Evaluate >_
static int32_t evalGreaterUnsigned(
    int32_t leftWord,uint32_t leftValue,int32_t rightWord,uint32_t rightValue){
    return leftWord && rightWord ? leftValue > rightValue : (
        (leftWord  ? (float) leftValue  : asFloat(leftValue )) >
        (rightWord ? (float) rightValue : asFloat(rightValue))
    );
}

// Evaluate <=
static int32_t evalLessEqualSigned(
    int32_t leftWord,int32_t leftValue,int32_t rightWord,int32_t rightValue) {
    return leftWord && rightWord ? leftValue <= rightValue : (
        (leftWord  ? (float) leftValue  : asFloat(leftValue )) <=
        (rightWord ? (float) rightValue : asFloat(rightValue))
    );
}

// Evaluate <=_
static int32_t evalLessEqualUnsigned(
    int32_t leftWord,uint32_t leftValue,int32_t rightWord,uint32_t rightValue){
    return leftWord && rightWord ? leftValue <= rightValue : (
        (leftWord  ? (float) leftValue  : asFloat(leftValue )) <=
        (rightWord ? (float) rightValue : asFloat(rightValue))
    );
}

// Evaluate <
static int32_t evalLessSigned(
    int32_t leftWord,int32_t leftValue,int32_t rightWord,int32_t rightValue) {
    return leftWord && rightWord ? leftValue < rightValue : (
        (leftWord  ? (float) leftValue  : asFloat(leftValue )) <
        (rightWord ? (float) rightValue : asFloat(rightValue))
    );
}

// Evaluate <_
static int32_t evalLessUnsigned(
    int32_t leftWord,uint32_t leftValue,int32_t rightWord,uint32_t rightValue){
    return leftWord && rightWord ? leftValue < rightValue : (
        (leftWord  ? (float) leftValue  : asFloat(leftValue )) <
        (rightWord ? (float) rightValue : asFloat(rightValue))
    );
}

// Evaluate *
static int32_t evalMultiply(
    int32_t leftWord,int32_t leftValue,int32_t rightWord,int32_t rightValue) {
    return leftWord && rightWord ? leftValue * rightValue : asWord(
        (leftWord  ? (float) leftValue  : asFloat(leftValue )) *
        (rightWord ? (float) rightValue : asFloat(rightValue))
    );
}

// Evaluate !=
static int32_t evalNotEqual(
    int32_t leftWord,int32_t leftValue,int32_t rightWord,int32_t rightValue) {
    return leftWord && rightWord ? leftValue != rightValue : (
        (leftWord  ? (float) leftValue  : asFloat(leftValue )) !=
        (rightWord ? (float) rightValue : asFloat(rightValue))
    );
}

// Evaluate %
static int32_t evalRemainder(
    int32_t leftWord,int32_t leftValue,int32_t rightWord,int32_t rightValue) {
    return !(rightWord ? rightValue : rightValue & 0x7FFFFFFF) ? 0 :
    leftWord && rightWord ? leftValue % rightValue : asWord(fmodf(
        (leftWord  ? (float) leftValue  : asFloat(leftValue )),
        (rightWord ? (float) rightValue : asFloat(rightValue))
    ));
}

// Evaluate <<
static int32_t evalShiftLeft(
    int32_t leftWord,int32_t leftValue,int32_t rightWord,int32_t rightValue) {
    leftValue =  leftWord  ? leftValue  : toWord(asFloat(leftValue ));
    leftValue = (rightWord ? rightValue : toWord(asFloat(rightValue))) & 31;
    return leftValue << rightValue;
}

// Evaluate >>
static int32_t evalShiftRightArithmetic(
    int32_t leftWord,int32_t leftValue,int32_t rightWord,int32_t rightValue) {
    leftValue =  leftWord  ? leftValue  : toWord(asFloat(leftValue ));
    leftValue = (rightWord ? rightValue : toWord(asFloat(rightValue))) & 31;
    return !rightValue ? leftValue :
        leftValue >> rightValue & (1 << 32 - rightValue) - 1;
}

// Evaluate >>>
static int32_t evalShiftRight(
    int32_t leftWord,int32_t leftValue,int32_t rightWord,int32_t rightValue) {
    leftValue =  leftWord  ? leftValue  : toWord(asFloat(leftValue ));
    leftValue = (rightWord ? rightValue : toWord(asFloat(rightValue))) & 31;
    if (!rightValue)
        return leftValue;
    leftValue >>=      rightValue;
    rightValue  = 32 - rightValue;
    return SIGN_EXTEND(rightValue, leftValue);
}

// Evaluate -
static int32_t evalSubtract(
    int32_t leftWord,int32_t leftValue,int32_t rightWord,int32_t rightValue) {
    return leftWord && rightWord ? leftValue - rightValue : asWord(
        (leftWord  ? (float) leftValue  : asFloat(leftValue )) -
        (rightWord ? (float) rightValue : asFloat(rightValue))
    );
}



///////////////////////////////////////////////////////////////////////////////
//                             Symbol Functions                              //
///////////////////////////////////////////////////////////////////////////////

// Evaluate address
static int32_t evalAddress(Vue *vue) {
    if (vue->cpu.inst.format == 6)
        return vue->cpu.program[vue->cpu.inst.reg1] + vue->cpu.inst.disp;
    switch (vue->cpu.inst.id) {
        case VUE_BCOND: case VUE_JAL: case VUE_JMP: case VUE_JR:
            return vue->cpu.pc + vue->cpu.inst.disp;
        case VUE_RETI:
            return vue->cpu.psw_np ? vue->cpu.fepc : vue->cpu.eipc;
        case VUE_TRAP:
            return 0xFFFFFFA0 + vue->cpu.inst.imm & 0xFFFFFFF0;
    }
    return 0;
}

// Evaluate cond
static int32_t evalCond(Vue *vue) {
    switch (vue->cpu.inst.id) {
        case VUE_BCOND: return vue->cpu.inst.cond;
        case VUE_SETF : return vue->cpu.inst.imm;
    }
    return 0;
}

// Evaluate disp
static int32_t evalDisp(Vue *vue) {
    switch (vue->cpu.inst.format) {
        case 3: case 4: case 6: return vue->cpu.inst.disp;
    }
    return 0;
}

// Evaluate imm
static int32_t evalImm(Vue *vue) {
    switch (vue->cpu.inst.format) {
        case 2: case 5: return vue->cpu.inst.imm;
    }
    return 0;
}

// Evaluate reg1
static int32_t evalReg1(Vue *vue) {
    switch (vue->cpu.inst.format) {
        case 1: case 5: case 6: case 7: return vue->cpu.inst.reg1;
    }
    return 0;
}

// Evaluate reg2
static int32_t evalReg2(Vue *vue) {
    switch (vue->cpu.inst.format) {
        case 1: case 2: case 5: case 6: case 7: return vue->cpu.inst.reg2;
    }
    return 0;
}

// Evaluate regid
static int32_t evalRegId(Vue *vue) {
    switch (vue->cpu.inst.id) {
        case VUE_LDSR: case VUE_STSR: return vue->cpu.inst.imm;
    }
    return 0;
}

// Evaluate subopcode
static int32_t evalSubopcode(Vue *vue) {
    switch (vue->cpu.inst.opcode) {
        case 0x1F: return vue->cpu.inst.imm;
        case 0x3E: return vue->cpu.inst.subopcode;
    }
    return 0;
}

// Evaluate value
static int evalValue(Vue *vue) {
    return vue->cpu.inst.format == 6 ? vue->cpu.access.value : 0;
}

// Evaluate vector
static int evalVector(Vue *vue) {
    return vue->cpu.inst.id == VUE_TRAP ? vue->cpu.inst.imm : 0;
}



///////////////////////////////////////////////////////////////////////////////
//                              Unary Functions                              //
///////////////////////////////////////////////////////////////////////////////

// Evaluate ~
static int32_t evalBitwiseNot(int32_t isWord, int32_t value) {
    return isWord ? ~value : ~toWord(asFloat(value));
}

// Evaluate !
static int32_t evalLogicalNot(int32_t isWord, int32_t value) {
    return isWord ? !value : !asFloat(value);
}

// Evaluate -
static int32_t evalNegate(int32_t isWord, int32_t value) {
    return isWord ? -value : asWord(-asFloat(value));
}

// Evaluate ceil
static int32_t evalCeil(int32_t isWord, int32_t value) {
    return isWord ? value : asWord(ceilf(asFloat(value)));
}

// Evaluate float
static int32_t evalFloat(int32_t isWord, int32_t value) {
    return isWord ? asWord((float) value) : value;
}

// Evaluate floor
static int32_t evalFloor(int32_t isWord, int32_t value) {
    return isWord ? value : asWord(floorf(asFloat(value)));
}

// Evaluate []
static int32_t evalReadWord(int32_t isWord, int32_t value, Vue *vue) {
    return vueRead(vue, isWord ? value : toWord(asFloat(value)), VUE_S32);
}

// Evaluate round
static int32_t evalRound(int32_t isWord, int32_t value) {
    return isWord ? value : asWord(roundf(asFloat(value)));
}

// Evaluate trunc
static int32_t evalTrunc(int32_t isWord, int32_t value) {
    return isWord ? value : asWord(truncf(asFloat(value)));
}

// Evaluate word
static int32_t evalWord(int32_t isWord, int32_t value) {
    return isWord ? value : toWord(asFloat(value));
}

// Evaluate xfloat
static int32_t evalXFloat(int32_t isWord, int32_t value) {
    return isWord ? asWord(asFloat(value)) : value;
}



///////////////////////////////////////////////////////////////////////////////
//                           Evaluation Functions                            //
///////////////////////////////////////////////////////////////////////////////

// Evaluate a binary operator
static int32_t evalBinary(int32_t id, int32_t *stack, int32_t size) {
    int32_t rv = stack[size - 1];
    int32_t rt = stack[size - 2];
    int32_t rw = rt == WORD;
    int32_t lv = stack[size - 3];
    int32_t lt = stack[size - 4];
    int32_t lw = lt == WORD;
    int32_t type  = rw && lw ? WORD : FLOAT;
    int32_t value = 0;
    switch (id) {

        // Arithmetic
        case ADD                   :
            value = evalAdd                 (lw, lv, rw, rv); break;
        case DIVIDE                :
            value = evalDivide              (lw, lv, rw, rv); break;
        case MULTIPLY              :
            value = evalMultiply            (lw, lv, rw, rv); break;
        case REMAINDER             :
            value = evalRemainder           (lw, lv, rw, rv); break;
        case SUBTRACT              :
            value = evalSubtract            (lw, lv, rw, rv); break;
        default:
            type = WORD; switch (id) {

        // Bitwise
        case BITWISE_AND           :
            value = evalBitwiseAnd          (lw, lv, rw, rv); break;
        case BITWISE_OR            :
            value = evalBitwiseOr           (lw, lv, rw, rv); break;
        case BITWISE_XOR           :
            value = evalBitwiseXOr          (lw, lv, rw, rv); break;
        case SHIFT_LEFT            :
            value = evalShiftLeft           (lw, lv, rw, rv); break;
        case SHIFT_RIGHT           :
            value = evalShiftRight          (lw, lv, rw, rv); break;
        case SHIFT_RIGHT_ARITHMETIC:
            value = evalShiftRightArithmetic(lw, lv, rw, rv); break;

        // Relational
        case EQUAL                 :
            value = evalEqual               (lw, lv, rw, rv); break;
        case GREATER_EQUAL_SIGNED  :
            value = evalGreaterEqualSigned  (lw, lv, rw, rv); break;
        case GREATER_EQUAL_UNSIGNED:
            value = evalGreaterEqualUnsigned(lw, lv, rw, rv); break;
        case GREATER_SIGNED        :
            value = evalGreaterSigned       (lw, lv, rw, rv); break;
        case GREATER_UNSIGNED      :
            value = evalGreaterUnsigned     (lw, lv, rw, rv); break;
        case LESS_EQUAL_SIGNED     :
            value = evalLessEqualSigned     (lw, lv, rw, rv); break;
        case LESS_EQUAL_UNSIGNED   :
            value = evalLessEqualUnsigned   (lw, lv, rw, rv); break;
        case LESS_SIGNED           :
            value = evalLessSigned          (lw, lv, rw, rv); break;
        case LESS_UNSIGNED         :
            value = evalLessUnsigned        (lw, lv, rw, rv); break;
        case NOT_EQUAL             :
            value = evalNotEqual            (lw, lv, rw, rv); break;

        // Logical
        default:
            lw = (lv & (lw ? 0xFFFFFFFF : 0x7FFFFFFF)) != 0;
            rw = (rv & (rw ? 0xFFFFFFFF : 0x7FFFFFFF)) != 0;

        // Evaluate &&
        case LOGICAL_AND:
            type = lt; value = lv;
            if (lw)
                { type = rt; value = rv; }
            break;

        // Evaluate ||
        case LOGICAL_OR:
            type = lt; value = lv;
            if (!lw)
                { type = rt; value = rv; }
            break;

        // Evaluate ^^
        case LOGICAL_XOR:
            type = rt; value = rv;
            if (lw == rw)
                { type = WORD; value = 0; }
            else if (lw)
                { type = lt; value = lv; }
            break;
    }}
    stack[size - 4] = type;
    stack[size - 3] = value;
    return size - 2;
}

// Evaluate a functional symbol
static int32_t evalSymbol(Vue *vue, int32_t id, int32_t *stack, int32_t size) {
    int32_t ret = 0;
    if (id == VUE_PC)
        ret = vue->cpu.pc;
    else if (id >= 200)
        ret = vueGetRegister(vue, id - 200, VUE_TRUE);
    else if (id >= 100)
        ret = vue->cpu.program[id - 100];
    else if (vue->cpu.stage == EXCEPTION && id == CODE)
        ret = vue->cpu.exception.code;
    else if (vue->cpu.stage == EXECUTE) switch (id) {
        case ADDRESS  : ret = evalAddress    (vue); break;
        case COND     : ret = evalCond       (vue); break;
        case DISP     : ret = evalDisp       (vue); break;
        case FORMAT   : ret = vue->cpu.inst.format; break;
        case ID       : ret = vue->cpu.inst.id    ; break;
        case IMM      : ret = evalImm        (vue); break;
        case OPCODE   : ret = vue->cpu.inst.opcode; break;
        case REG1     : ret = evalReg1       (vue); break;
        case REG2     : ret = evalReg2       (vue); break;
        case REGID    : ret = evalRegId      (vue); break;
        case SIZE     : ret = vue->cpu.inst.size  ; break;
        case SUBOPCODE: ret = evalSubopcode  (vue); break;
        case VALUE    : ret = evalValue      (vue); break;
        case VECTOR   : ret = evalVector     (vue); break;
    }
    stack[size++] = WORD;
    stack[size++] = ret;
    return size;
}

// Evaluate a unary operator
static void evalUnary(Vue *vue, int32_t id, int32_t *stack, int32_t size) {
    int32_t type   = stack[size - 2];
    int32_t value  = stack[size - 1];
    int32_t isWord = type == WORD;
    switch (id) {
        case BITWISE_NOT: value = evalBitwiseNot(isWord, value);
            type = WORD ; break;
        case LOGICAL_NOT: value = evalLogicalNot(isWord, value);
            type = WORD ; break;
        case NEGATE     : value = evalNegate    (isWord, value); break;
        case CEIL       : value = evalCeil      (isWord, value); break;
        case FLOAT      : value = evalFloat     (isWord, value);
            type = FLOAT; break;
        case FLOOR      : value = evalFloor     (isWord, value); break;
        case READ_WORD  : value = evalReadWord  (isWord, value, vue);
            type = WORD ; break;
        case ROUND      : value = evalRound     (isWord, value); break;
        case TRUNC      : value = evalTrunc     (isWord, value); break;
        case WORD       : value = evalWord      (isWord, value);
            type = WORD ; break;
        case XFLOAT     : value = evalXFloat    (isWord, value);
            type = FLOAT; break;
        case XWORD      : type = WORD;
    }
    stack[size - 2] = type;
    stack[size - 1] = value;
}

// Evaluate a breakpoint condition for an emulation context
int32_t evaluate(Vue *vue, Breakpoint *brk, int32_t *stack) {

    // The condition is empty
    if (brk->numCondition == 0)
        return brk->enabled;

    // Process tokens
    int size = 0;
    for (int x = 0; x < brk->numCondition; x++) {
        Token *tok = &brk->condition[x];
        switch (tok->type) {
            case BINARY: size =
                evalBinary(     tok->value, stack, size); continue;
            case SYMBOL: size =
                evalSymbol(vue, tok->value, stack, size); continue;
            case UNARY :
                evalUnary (vue, tok->value, stack, size); continue;
        }
        stack[size++] = tok->type;
        stack[size++] = tok->value;
    }
    return (stack[1] & (stack[0] == WORD ? 0xFFFFFFFF : 0x7FFFFFFF)) != 0;
}

#endif // NATIVEVUE
